#ifdef __APPLE__
#include "GLUT/glut.h"
#include <OPENGL/gl.h>
#endif
#ifdef __unix__
#include <GL/glut.h>
#endif

#include <fstream>

#include "CMeshLoaderSimple.h" //For .o files
#include <stdlib.h>
#include <stdio.h>
#include "GLSL_helper.h"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp" //perspective, trans etc
#include "glm/gtc/type_ptr.hpp" //value_ptr
#include "RenderingHelper.h" //For ModelTrans
#include "GeometryCreator.h" //Create VBO's of basic shapes
#include "particle.h" //my particle generator

#include "quat.h" //Quaternion Library

#include "mesh.h" //Ross' mesh loader for .obj files

using namespace std;
using namespace glm;

#define PI 3.141592654

#define MESH_GROUND 0
#define MESH_BUNNY 1
#define MESH_DRAGON 2
#define MESH_BONE 3
#define MESH_SUIT 4
#define MESH_SUIT_LIGHT 5
#define MESH_MOON 6
#define MESH_PUMPKIN 7
#define MESH_RAINBOW 8

#define SCALE_JACK .03f
#define NUM_QUATS 8

typedef struct {
  GLuint VBO;
  GLuint IBO;
  GLuint NBO;
  int IBOtris;
  vec3 toJoint;
  vec3 toBody;
  mat4 rotation;
  Quat quats[NUM_QUATS];
  int mesh;
} Jack;

void initJackPart (Jack &j) {
  j.rotation = mat4(1.0f);
  j.toJoint = vec3(0, 0, 0);
  j.toBody = vec3(0, 0, 0);
  for (int i = 0; i < NUM_QUATS; i++) { j.quats[i] = Quat::makeQuat(0, Cvec3(0, 0, 1)); }
  j.mesh = MESH_SUIT;
}


#define NUM_JACK_PARTS 9
Jack j_Body, j_Head, j_UpperArm, j_LowerArm, j_Hand, j_RUpperLeg, j_LUpperLeg, j_LLowerLeg, j_RLowerLeg;

//used to iterate through all the parts for updating animation
Jack *j_parts[NUM_JACK_PARTS] = {&j_Body, &j_Head, &j_UpperArm, &j_LowerArm, &j_Hand, &j_RUpperLeg, &j_LUpperLeg, &j_LLowerLeg, &j_RLowerLeg};

#define NUM_PUMPKINS 225
vec3 pumpkins[NUM_PUMPKINS];
int pumpkinsRot[NUM_PUMPKINS];

//flag and ID to toggle on and off the shader
int shade = 1;
int ShadeProg;

//Handles to the shader data
GLint h_aPosition;
GLint h_aNormal;
        
GLint h_uProjMatrix;
GLint h_uViewMatrix;
GLint h_uModelMatrix;
  	  
GLint h_uLightPos;
GLint h_uLightColor;
GLint h_uLightDir;
GLint h_uBrightness;
        
GLint h_uMatAmb;
GLint h_uMatDif;
GLint h_uMatSpec;
GLint h_uMatShine;
GLint h_uMatAlpha;

GLuint GrndBuffObj, GIndxBuffObj, GroundNormalBO;
unsigned int g_CiboLen, g_GiboLen;

GLuint BunnyIBO, BunnyVBO, BunnyNormalBO;
unsigned int BunnyTriangleCount;

GCMesh* Cube;
GCMesh* Sphere;
GCMesh* FirePart;

vec3 g_wind = vec3(.5, 0, 1);
int g_enableWind = 0;

RenderingHelper ModelTrans;

static float g_width, g_height;

float g_lastAnimateTime = 0;
float g_resetTime = 0;

float g_cameraSpeed = 1.0f;
float g_camX = 8.0;
float g_camY = 5.0;
float g_camZ = 8.34;
float g_camLookX = -0.92;
float g_camLookY = -0.22;
float g_camLookZ = -0.31;
float g_camTheta = -2.274819;
float g_camPhi = -.133290;

float g_brightness = -0.2;
bool g_showPath = false;

float g_mouseX = 0;
float g_mouseY = 0;

float g_transX = 0;
float g_transY = 0;
float g_transZ = 0;


float g_JackX = 0;
float g_JackY = 0;
float g_JackZ = 0;
float g_quatRot = 0;
mat4 g_JackBodyRot = mat4();

static const float g_groundY = -1.0;      // y coordinate of the ground
static const float g_groundSize = 120.0;   // half the ground length

#define NUM_POSITIONS 14
vec3 positions[NUM_POSITIONS] = 
          { vec3(0, 0, 0),
            vec3(0, 0, 1),
            vec3(0, 0, 2),
            vec3(0, 0, 3),
            vec3(0, 0, 4),
            vec3(0, 0, 5),
            vec3(0, 0, 6),
            vec3(1 ,0, 6),
            vec3(2, 0, 5),
            vec3(2, 0, 4),
            vec3(2, 0, 3),
            vec3(2, 0, 2),
            vec3(2, 0, 1),
            vec3(2, 0, 0),
          };

Quat rotations[NUM_POSITIONS] = {
            Quat::makeQuat(0, Cvec3(0, 1, 0)),
            Quat::makeQuat(0, Cvec3(0, 1, 0)),
            Quat::makeQuat(0, Cvec3(0, 1, 0)),
            Quat::makeQuat(0, Cvec3(0, 1, 0)),
            Quat::makeQuat(0, Cvec3(0, 1, 0)),
            Quat::makeQuat(0, Cvec3(0, 1, 0)),
            Quat::makeQuat(0, Cvec3(0, 1, 0)),
            Quat::makeQuat(80, Cvec3(0, 1, 0)),
            Quat::makeQuat(180, Cvec3(0, 1, 0)),
            Quat::makeQuat(180, Cvec3(0, 1, 0)),
            Quat::makeQuat(180, Cvec3(0, 1, 0)),
            Quat::makeQuat(180, Cvec3(0, 1, 0)),
            Quat::makeQuat(180, Cvec3(0, 1, 0)),
            Quat::makeQuat(180, Cvec3(0, 1, 0)),
            };

/* projection matrix */
void SetProjectionMatrix() {
  glm::mat4 Projection = glm::perspective(80.0f, (float)g_width/g_height, 0.1f, 2000.f);	
  safe_glUniformMatrix4fv(h_uProjMatrix, glm::value_ptr(Projection));
}

/* camera controls */
void SetView() {   
  mat4 Camera = lookAt(vec3(g_camX, g_camY, g_camZ), 
                       vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ),
                       vec3( 0.0, 1.0, 0.0));
                       
  //printf("Camera: at(%f, %f, %f), looking at(%f, %f, %f)\n", 
  //         g_camX, g_camY, g_camZ, g_camLookX+g_camX, g_camY+g_camLookY, g_camZ+g_camLookZ);
                       
  safe_glUniformMatrix4fv(h_uViewMatrix, glm::value_ptr(Camera));
  glutWarpPointer(g_width/2, g_height/2);
  glutPostRedisplay();

}

void SetModel() {
  safe_glUniformMatrix4fv(h_uModelMatrix, glm::value_ptr(ModelTrans.modelViewMatrix));
}

/* intialize ground data */
static void initGround() {

  // A x-z plane at y = g_groundY of dimension [-g_groundSize, g_groundSize]^2
    float GrndPos[] = { 
    -g_groundSize, g_groundY, -g_groundSize, 
    -g_groundSize, g_groundY,  g_groundSize, 
     g_groundSize, g_groundY,  g_groundSize, 
     g_groundSize, g_groundY, -g_groundSize
    };
    
    float GrndNorm[] = {0, 1, 0, 0, 1, 0};
    unsigned short idx[] = {0, 1, 2, 0, 2, 3};
  
    g_GiboLen = 6;
    glGenBuffers(1, &GrndBuffObj);
    glBindBuffer(GL_ARRAY_BUFFER, GrndBuffObj);
    glBufferData(GL_ARRAY_BUFFER, sizeof(GrndPos), GrndPos, GL_STATIC_DRAW);
    
    glGenBuffers(1, &GroundNormalBO);
    glBindBuffer(GL_ARRAY_BUFFER, GroundNormalBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(GrndNorm), GrndNorm, GL_STATIC_DRAW);

    glGenBuffers(1, &GIndxBuffObj);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, GIndxBuffObj);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);
}

void copyDrawData(Jack &j1, Jack &j2) {
  j2.IBOtris = j1.IBOtris;
  j2.VBO = j1.VBO;
  j2.IBO = j1.IBO;
  j2.NBO = j1.NBO;
  j2.toJoint = j1.toJoint;
  j2.toBody = j1.toBody;
  j2.mesh = j1.mesh;
}

void initJack() {
  // Defaults most values and vecs to 0 
  for (int i = 0; i < NUM_JACK_PARTS; i++) {
    initJackPart(*j_parts[i]);
  }

  // open an OBJ file
  std::ifstream modelFile("Models/skellington.obj");
  // parse the meshes from the file
  Model model;
  model.load(modelFile);

  // You can get fancy and check against mesh names:
  for (MeshVector::const_iterator iter = model.meshes().begin(); iter != model.meshes().end(); ++iter) {
    if (iter->name() == "Head") {
      j_Head.IBOtris = iter->makeVBO(&j_Head.IBO, &j_Head.VBO, NULL, &j_Head.NBO);
    } else if (iter->name() == "Body") {
      j_Body.IBOtris = iter->makeVBO(&j_Body.IBO, &j_Body.VBO, NULL, &j_Body.NBO);
    } else if (iter->name() == "UpperArm") {
      j_UpperArm.IBOtris = iter->makeVBO(&j_UpperArm.IBO, &j_UpperArm.VBO, NULL, &j_UpperArm.NBO);
    } else if (iter->name() == "LowerArm") {
      j_LowerArm.IBOtris = iter->makeVBO(&j_LowerArm.IBO, &j_LowerArm.VBO, NULL, &j_LowerArm.NBO);
    } else if (iter->name() == "Hand") {
      j_Hand.IBOtris = iter->makeVBO(&j_Hand.IBO, &j_Hand.VBO, NULL, &j_Hand.NBO);
    } else if (iter->name() == "UpperLeg") {
      j_LUpperLeg.IBOtris = iter->makeVBO(&j_LUpperLeg.IBO, &j_LUpperLeg.VBO, NULL, &j_LUpperLeg.NBO);
    } else if (iter->name() == "LowerLeg") {
      j_LLowerLeg.IBOtris = iter->makeVBO(&j_LLowerLeg.IBO, &j_LLowerLeg.VBO, NULL, &j_LLowerLeg.NBO);
    }
  }


  #include "jack_animation.h"

  j_Head.toBody = vec3(0, 1.5, .01) / SCALE_JACK;
  j_Head.toJoint = vec3(0, 0, 0) / SCALE_JACK;
  j_Head.mesh = MESH_BONE;

  j_UpperArm.toJoint = vec3(0.53, 0.0, 0.0) / SCALE_JACK;
  j_UpperArm.toBody = vec3(.9, .7, 0) / SCALE_JACK;

  j_LowerArm.toJoint = vec3(.78, .01, 0.0) / SCALE_JACK;
  j_LowerArm.toBody = vec3(1.61, 0, 0) / SCALE_JACK;

  j_Hand.toJoint = vec3(.39, -.01, 0) / SCALE_JACK;
  j_Hand.toBody = vec3(.93, .02, 0.0) / SCALE_JACK;
  j_Hand.mesh = MESH_BONE;

  j_LUpperLeg.toJoint = vec3(.01, -.84, 0) / SCALE_JACK;
  j_LUpperLeg.toBody = vec3(.15, -1.49, 0) / SCALE_JACK;

  j_LLowerLeg.toJoint = vec3(0, -1.88, -.08) / SCALE_JACK;
  j_LLowerLeg.toBody = vec3(.01, -3.03, .01) / SCALE_JACK;
  
  copyDrawData(j_LLowerLeg, j_RLowerLeg);
  copyDrawData(j_LUpperLeg, j_RUpperLeg);
}

// ASSUME 225 PUMPKINS!!!
void initPumpkins() {
  int count = 0;
  for (int i = -15; i < 15; i+=2) {
    for (int j = -15; j < 15; j+=2) {
      pumpkins[count] = vec3((rand() % 100) / 100.0f + (i*2), 0, (rand() % 100) / 100.0f + (j*2));
      pumpkinsRot[count] = rand() % 360;
      count++;
    }
  }
}

void InitGeom() {
  initGround();
  initJack();
  initPumpkins();
  initParticles();
  CMeshLoader::loadVertexBufferObjectFromMesh("Models/bunny500.m", BunnyTriangleCount, BunnyVBO, BunnyIBO, BunnyNormalBO);
  //CMeshLoader::loadVertexBufferObjectFromMesh("Models/bunny500.m", Bunny->IndexBufferLength, Bunny->PositionHandle, Bunny->IndexHandle, Bunny->NormalHandle);

  Cube = GeometryCreator::CreateCube();

  //vec3(radius), slices, stacks
  FirePart = GeometryCreator::CreateSphere(vec3(.1, .05, .2), 3, 2);
  Sphere = GeometryCreator::CreateSphere();
}

/*function to help load the shaders (both vertex and fragment */
int InstallShader(const GLchar *vShaderName, const GLchar *fShaderName) {
        GLuint VS; //handles to shader object
        GLuint FS; //handles to frag shader object
        GLint vCompiled, fCompiled, linked; //status of shader

        VS = glCreateShader(GL_VERTEX_SHADER);
        FS = glCreateShader(GL_FRAGMENT_SHADER);

        //load the source
        glShaderSource(VS, 1, &vShaderName, NULL);
        glShaderSource(FS, 1, &fShaderName, NULL);

        //compile shader and print log
        glCompileShader(VS);
        /* check shader status requires helper functions */
        printOpenGLError();
        glGetShaderiv(VS, GL_COMPILE_STATUS, &vCompiled);
        printShaderInfoLog(VS);

        //compile shader and print log
        glCompileShader(FS);
        /* check shader status requires helper functions */
        printOpenGLError();
        glGetShaderiv(FS, GL_COMPILE_STATUS, &fCompiled);
        printShaderInfoLog(FS);

        if (!vCompiled || !fCompiled) {
                printf("Error compiling either shader %s or %s", vShaderName, fShaderName);
                return 0;
        }

        //create a program object and attach the compiled shader
        ShadeProg = glCreateProgram();
        glAttachShader(ShadeProg, VS);
        glAttachShader(ShadeProg, FS);

        glLinkProgram(ShadeProg);
        /* check shader status requires helper functions */
        printOpenGLError();
        glGetProgramiv(ShadeProg, GL_LINK_STATUS, &linked);
        printProgramInfoLog(ShadeProg);

        glUseProgram(ShadeProg);

        /* get handles to attribute data */
        h_aPosition = safe_glGetAttribLocation(ShadeProg, "aPosition");
        h_aNormal = safe_glGetAttribLocation(ShadeProg, "aNormal");
        
    	  h_uProjMatrix = safe_glGetUniformLocation(ShadeProg, "uProjMatrix");
    	  h_uViewMatrix = safe_glGetUniformLocation(ShadeProg, "uViewMatrix");
    	  h_uModelMatrix = safe_glGetUniformLocation(ShadeProg, "uModelMatrix");
    	  
    	  h_uLightPos = safe_glGetUniformLocation(ShadeProg, "uLightPos");
        h_uLightColor = safe_glGetUniformLocation(ShadeProg, "uLightColor");
        h_uBrightness = safe_glGetUniformLocation(ShadeProg, "uBrightness");
        //h_uLightDir = safe_glGetUniformLocation(ShadeProg, "uLightDir");
        
        h_uMatAmb = safe_glGetUniformLocation(ShadeProg, "uMat.aColor");
        h_uMatDif = safe_glGetUniformLocation(ShadeProg, "uMat.dColor");
        h_uMatSpec = safe_glGetUniformLocation(ShadeProg, "uMat.sColor");
        h_uMatShine = safe_glGetUniformLocation(ShadeProg, "uMat.shine");
        h_uMatAlpha = safe_glGetUniformLocation(ShadeProg, "uMat.alpha");


        printf("sucessfully installed shader %d\n", ShadeProg);
        return 1;
	
}


/* Some OpenGL initialization */
void Initialize ()
{
	// Start Of User Initialization
	glClearColor (14/255.0f, 19/255.0f, 38/255.0f, 1.0f);		
   ModelTrans.useModelViewMatrix();
   ModelTrans.loadIdentity();						
	// Black Background
 	glClearDepth (1.0f);	// Depth Buffer Setup
 	glDepthFunc (GL_LEQUAL);	// The Type Of Depth Testing
	glEnable (GL_DEPTH_TEST);// Enable Depth Testing
  glEnable (GL_BLEND); // for transparency
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

}

void SetMesh(int s) {
   glUniform1f(h_uMatAlpha, 1.0);
   switch (s) {
   case MESH_GROUND:
      glUniform3f(h_uMatAmb, 0.2, 0.2, 0.2);
      glUniform3f(h_uMatDif, 48/255.0, 34/255.0, 23/255.0);
      glUniform3f(h_uMatSpec, 0.3, 0.5, 0.3);
      glUniform1f(h_uMatShine, 1.0); 
      break;
   case MESH_BUNNY:
      glUniform3f(h_uMatAmb, 0.2, 0.2, 0.2);
      glUniform3f(h_uMatDif, .7, .1, .7);
      glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1	);
      glUniform1f(h_uMatShine, 5.0);
      break;
   case MESH_DRAGON: 
      glUniform3f(h_uMatAmb, .15, .45, .15);
      glUniform3f(h_uMatDif, .3, .9, .3);
      glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1	);
      glUniform1f(h_uMatShine, .01);
      break;
   case MESH_BONE:
      glUniform3f(h_uMatAmb, .3, .3, .3);
      glUniform3f(h_uMatDif, .95, .95, .95);
      glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1	);
      glUniform1f(h_uMatShine, .01);
      break;
   case MESH_SUIT:
      glUniform3f(h_uMatAmb, .1, .1, .1);
      glUniform3f(h_uMatDif, .2, .2, .2);
      glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1	);
      glUniform1f(h_uMatShine, .01);
      break;
   case MESH_SUIT_LIGHT:
      glUniform3f(h_uMatAmb, .15, .15, .15);
      glUniform3f(h_uMatDif, .3, .3, .3);
      glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1	);
      glUniform1f(h_uMatShine, .01);
      break;
   case MESH_MOON:
      glUniform3f(h_uMatAmb, .9, .9, .3);
      glUniform3f(h_uMatDif, 1.0, 1.0, 0.0);
      glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1	);
      glUniform1f(h_uMatShine, .01);
      break;
   case MESH_PUMPKIN:
     glUniform3f(h_uMatAmb, .3, .1, .0);
     glUniform3f(h_uMatDif, 232.0/255.0, 97.0/255.0, 0.0);
     glUniform3f(h_uMatSpec, 0.2, 0.1, 0.0	);
     glUniform1f(h_uMatShine, .1);
     break;
      
   }
}

void DrawBunny() {
  SetModel();
  SetMesh(MESH_BUNNY);

  safe_glEnableVertexAttribArray(h_aPosition);
 	glBindBuffer(GL_ARRAY_BUFFER, BunnyVBO);
	safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
	
	safe_glEnableVertexAttribArray(h_aNormal);
 	glBindBuffer(GL_ARRAY_BUFFER, BunnyNormalBO);
	safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, BunnyIBO);
  glDrawElements(GL_TRIANGLES, BunnyTriangleCount*3, GL_UNSIGNED_INT, 0);

	//clean up 
	safe_glDisableVertexAttribArray(h_aPosition);
	safe_glDisableVertexAttribArray(h_aNormal);
}

void DrawGround() {
  SetModel();
  SetMesh(MESH_GROUND);

  safe_glEnableVertexAttribArray(h_aPosition);
  glBindBuffer(GL_ARRAY_BUFFER, GrndBuffObj);
  safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

  safe_glEnableVertexAttribArray(h_aNormal);
 	glBindBuffer(GL_ARRAY_BUFFER, GroundNormalBO);
	safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, GIndxBuffObj);
  glDrawElements(GL_TRIANGLES, g_GiboLen, GL_UNSIGNED_SHORT, 0);;

  	
  // Disable the attributes used by our shader
  safe_glDisableVertexAttribArray(h_aPosition);
  safe_glEnableVertexAttribArray(h_aPosition);
}

void DrawGCMesh(GCMesh* mesh) {
  SetModel();

  safe_glEnableVertexAttribArray(h_aPosition);
  glBindBuffer(GL_ARRAY_BUFFER, mesh->PositionHandle);
  safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

  safe_glEnableVertexAttribArray(h_aNormal);
  glBindBuffer(GL_ARRAY_BUFFER, mesh->NormalHandle);
  safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh->IndexHandle);

  glDrawElements(GL_TRIANGLES, mesh->IndexBufferLength, GL_UNSIGNED_SHORT, 0);

  safe_glDisableVertexAttribArray(h_aPosition);
  safe_glDisableVertexAttribArray(h_aNormal);
}

void DrawPumpkin() {
  SetMesh(MESH_PUMPKIN);
  ModelTrans.pushMatrix();
    ModelTrans.scale(.6, .7, 1.0);
    SetModel();
    DrawGCMesh(Sphere);
    ModelTrans.popMatrix();
  ModelTrans.pushMatrix();
    ModelTrans.scale(1.0, .7, .6);
    SetModel();
    DrawGCMesh(Sphere);
    ModelTrans.popMatrix();
  ModelTrans.pushMatrix();
    ModelTrans.translate(vec3(0.0, .7, 0.0));
    ModelTrans.scale(.1, .4, .1);
    SetMesh(MESH_DRAGON);
    SetModel();
    DrawGCMesh(Cube);
    ModelTrans.popMatrix();
}

void DrawJackPart(Jack &j) {
  SetModel();

  safe_glEnableVertexAttribArray(h_aPosition);
 	glBindBuffer(GL_ARRAY_BUFFER, j.VBO);
	safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
	
	safe_glEnableVertexAttribArray(h_aNormal);
 	glBindBuffer(GL_ARRAY_BUFFER, j.NBO);
	safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, j.IBO);
  glDrawElements(GL_TRIANGLES, j.IBOtris*3, GL_UNSIGNED_INT, 0);

	//clean up 
	safe_glDisableVertexAttribArray(h_aPosition);
	safe_glDisableVertexAttribArray(h_aNormal);
}

void ModelAndDrawPart(Jack &j, bool negativeX, bool rotate180) {
  if (negativeX) {
    ModelTrans.translate(j.toBody * vec3(-1, 1, 1)); // translate the other way
  } else {
    ModelTrans.translate(j.toBody);
  }
  if (rotate180) {
    ModelTrans.rotate(180, vec3(0, 1, 0)); // rotate since our .obj is the left side
  }
  ModelTrans.translate(j.toJoint * -1.0f);
  ModelTrans.multMatrix(j.rotation);
  ModelTrans.translate(j.toJoint);
  SetMesh(j.mesh);
  DrawJackPart(j);
}

/* Main display function */
void Draw (void) {

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);					
  glUseProgram(ShadeProg);

  //Set up the projection and view matricies
  SetProjectionMatrix();
  SetView();
   
  // Set up the light position, color, and direction
  glUniform3f(h_uLightPos, 0.0, 20.0, 10.0);
  //glUniform3f(h_uLightDir, g_camLookX, g_camLookY, g_camLookZ);
  glUniform3f(h_uLightColor, 1.0, 1.0, 1.0);
  glUniform3f(h_uBrightness, g_brightness, g_brightness, g_brightness);

  // Draw the Ground    
  DrawGround();

  // Draw the yellow moon - jack face!
  ModelTrans.pushMatrix();
    ModelTrans.translate(vec3(0.0, 20.0, -80.0)); // far off behind us 0, 20, -80
    ModelTrans.scale(2.0, 2.0, 2.0); 
    ModelTrans.rotate(180, vec3(0, 1, 0));
    SetModel();
    SetMesh(MESH_MOON);
    DrawJackPart(j_Head);
    //then move the light back to where it was
    ModelTrans.popMatrix();
  
  // Draw a mass of pumpkins 
  for (int i = 0; i < NUM_PUMPKINS; i++) {
    ModelTrans.pushMatrix();
      ModelTrans.translate(pumpkins[i]);
      ModelTrans.rotate(pumpkinsRot[i], vec3(0, 1, 0));
      DrawPumpkin();
    ModelTrans.popMatrix();
  }
  

  // Draw the particles
  Particle* p = getParticleArray();
  for (int i = 0; i < MAX_PARTICLES; i++) {
    ModelTrans.pushMatrix();
      float head_radius = .35;
      float tx = p->pos.x;
      float ty = p->pos.y;
      float tz = p->pos.z;

      // check if tx is inside the head
      if (tx < head_radius && tx > 0) { 
        tx = head_radius;
      } else if (tx > -1 * head_radius && tx < 0) {
        tx = -1*head_radius;
      }
      
      // check if tz is inside the head
      if (tz < head_radius && tz > 0) { 
        tz = head_radius;
      } else if (tz > -1 * head_radius && tz < 0) {
        tz = -1*head_radius;
      }

      ModelTrans.translate(vec3(tx, ty, tz));
      ModelTrans.scale(1.0f);
      ModelTrans.rotate((float)(rand() % 45), vec3(0, 1, 1));
      SetModel();
      glUniform3f(h_uMatAmb, p->color.x, p->color.y, p->color.z);
      glUniform3f(h_uMatDif, p->color.x, p->color.y, p->color.z);
      glUniform3f(h_uMatSpec, 0.0, 0.0, 0.0	);
      glUniform1f(h_uMatShine, .1);
      glUniform1f(h_uMatAlpha, p->color.w);
      DrawGCMesh(FirePart);
      //printf("part drawn: %f, %f, %f\n", p->pos.x, p->pos.y, p->pos.z);
    ModelTrans.popMatrix();
    p++;
  }



  //Draw our main character 
  //Draw the torso
  ModelTrans.pushMatrix();
    ModelTrans.translate(vec3(g_JackX, g_JackY, g_JackZ));
    ModelTrans.translate(vec3(0, 3.93, 0)); //how tall he stands (torso to ground distance
    ModelTrans.multMatrix(g_JackBodyRot);
    ModelTrans.scale(SCALE_JACK); //Our initial scale to reduce his entire body size since our obj is huge
    SetMesh(MESH_SUIT_LIGHT);
	  DrawJackPart(j_Body);

    //Draw the head
    ModelTrans.pushMatrix();
      ModelAndDrawPart(j_Head, false, false);
    ModelTrans.popMatrix();
    //pop the head

    ModelTrans.pushMatrix();
      ModelAndDrawPart(j_UpperArm, false, false);
      ModelTrans.pushMatrix(); 
        ModelAndDrawPart(j_LowerArm, false, false);
        ModelTrans.pushMatrix();
          ModelAndDrawPart(j_Hand, false, false);
        ModelTrans.popMatrix(); //hand
      ModelTrans.popMatrix(); //lower arm
    ModelTrans.popMatrix();  //upper arm

    ModelTrans.pushMatrix();
      ModelAndDrawPart(j_UpperArm, true, true);
      ModelTrans.pushMatrix();
        ModelAndDrawPart(j_LowerArm, false, false);
        ModelTrans.pushMatrix();
          ModelAndDrawPart(j_Hand, false, false);
        ModelTrans.popMatrix(); //hand
      ModelTrans.popMatrix(); //lower arm
    ModelTrans.popMatrix(); //upper arm

    ModelTrans.pushMatrix();
      ModelAndDrawPart(j_LUpperLeg, false, false);
      ModelTrans.pushMatrix();
        ModelAndDrawPart(j_LLowerLeg, false, false);
      ModelTrans.popMatrix(); //left lower leg
    ModelTrans.popMatrix(); //left upper leg

    //Draw the right leg
    ModelTrans.pushMatrix();
      ModelAndDrawPart(j_RUpperLeg, true, false);
      //Draw the right lower leg and foot
      ModelTrans.pushMatrix();
        ModelAndDrawPart(j_RLowerLeg, false, false);
      ModelTrans.popMatrix();
    ModelTrans.popMatrix();

  ModelTrans.popMatrix();
  //pop the torso
  

  // Draws our points along our path
  if (g_showPath) {
    for (int i = 0; i < NUM_POSITIONS; i++) {
      ModelTrans.pushMatrix();
        ModelTrans.translate(positions[i]);
        ModelTrans.scale(.4, .4, .4);
        DrawBunny();
      ModelTrans.popMatrix();
    }
  }
 

	//Disable the shader
	glUseProgram(0);	
	glutSwapBuffers();

}

vec3 interpTwoPoints(vec3 p0, vec3 p1, vec3 p2, vec3 p3, float time) {
  float t = time - (int)time;
   
  float t2 = t*t;
  float t3 = t*t*t;
   
  vec3 v1 = -1.0f*p0 +  3.0f*p1 + -3.0f*p2 +       p3;
  vec3 v2 =  2.0f*p0 + -5.0f*p1 +  4.0f*p2 + -1.0f*p3;
  vec3 v3 = -1.0f*p0 +  0.0f*p1 +       p2 +  0.0f*p3;
  vec3 v4 =  0.0f*p0 +  2.0f*p1 +  0.0f*p2 +  0.0f*p3;

  return (t3*v1 + t2*v2 + t*v3 + v4) * 0.5f; 
   
   
  //return alpha * p1 + (1.0f-alpha) * p0;
}

Quat lerpQuat(Quat q1, Quat q2, double s) {
  //printf("%f : angle=%f vec=<%f, %f, %f>\n", s, q1[0], q1[1], q1[2], q1[3]);
  //printf("%f : angle=%f vec=<%f, %f, %f>\n\n", s, q2[0], q2[1], q2[2], q2[3]);
  Quat q = q1*(1.0-s) + q2*(s);
  return q;

}

float myrand() {
  float x = 10000;
  return ((rand() % (int)x) / (x/2.0)) -1.0;
}

void animate(int trash) {
 
  long timeMs = glutGet(GLUT_ELAPSED_TIME) - g_resetTime;
  float timeS = timeMs / 1000.0;
  float timeElapsed = timeS - g_lastAnimateTime;
  g_lastAnimateTime = timeS;

  //Take care of our particle system updates
  for (int i = 0; i < 60; i++) {
    //addParticle(position, velocity)
    addParticle(glm::vec3((myrand()+1)/5.0 + g_JackX, (myrand()/5.0f) + 5.57f, (myrand()+1)/5.0 + g_JackZ),  
                glm::vec3(myrand()/10.0, (myrand()+3)/8.0, (myrand()+1)/10.0));
  }
  if (g_enableWind) {
    updateParticles(timeElapsed, g_wind);
  } else {
    updateParticles(timeElapsed, vec3(0, 0, 0));
  }

  glutTimerFunc(30.0, animate, 0);
  glutPostRedisplay();

  if (timeS > NUM_POSITIONS-2) {
    return;
  }
  //timeS += 1;
  vec3 newPosition = interpTwoPoints(positions[(int)std::max(timeS-1, 0.0f)], 
                                      positions[(int)timeS], 
                                      positions[(int)std::min(timeS+1, NUM_POSITIONS-1.0f)], 
                                      positions[(int)std::min(timeS+2, NUM_POSITIONS-1.0f)], 
                                      timeS
                                      );   
   
   
  g_JackX = newPosition.x;
  g_JackY = newPosition.y;
  g_JackZ = newPosition.z;

  
  g_JackBodyRot = quatToMatrix(lerpQuat(
                                     rotations[(int)timeS],
                                     rotations[(int)timeS+1],
                                     timeS - (int)timeS
                                    )
                           );
  
  
  for (int i = 0; i < NUM_JACK_PARTS; i++) {
    j_parts[i]->rotation = quatToMatrix(lerpQuat(
                                         j_parts[i]->quats[(int)timeS % NUM_QUATS],
                                         j_parts[i]->quats[((int)timeS+1) % NUM_QUATS],
                                         timeS - (int)timeS));
  }
         
}



/* Reshape */
void ReshapeGL (int width, int height)								
{
	g_width = (float)width;
	g_height = (float)height;
	glViewport (0, 0, (GLsizei)(width), (GLsizei)(height));				

}

//the keyboard callback to change the values to the transforms
void keyboard(unsigned char key, int x, int y ){
  switch( key ) {
    /* WASD keyes effect view/camera transform */
    case 'w': {
      vec3 la = vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ);
      vec3 eye = vec3(g_camX, g_camY, g_camZ);
      vec3 w = la - eye;
      w = normalize(w);
      eye = eye + g_cameraSpeed*w;
      g_camX = eye.x;
      g_camY = eye.y;
      g_camZ = eye.z;
      break;
    }
    case 's': {
      vec3 la = vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ);
      vec3 eye = vec3(g_camX, g_camY, g_camZ);
      vec3 w = la - eye;
      w = normalize(w);
      eye = eye + -g_cameraSpeed*w;
      g_camX = eye.x;
      g_camY = eye.y;
      g_camZ = eye.z;
      break;
    }	
    case 'a': {
      vec3 la = vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ);
      vec3 eye = vec3(g_camX, g_camY, g_camZ);
      vec3 w = la - eye;
      w = normalize(w);
      vec3 u = cross(vec3(0, 1, 0), w);
      u = normalize(u);      
      eye = eye + g_cameraSpeed*u;
      g_camX = eye.x;
      g_camY = eye.y;
      g_camZ = eye.z;
      break;
    }
    case 'd': {
      vec3 la = vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ);
      vec3 eye = vec3(g_camX, g_camY, g_camZ);
      vec3 w = la - eye;
      w = normalize(w);
      vec3 u = cross(vec3(0, 1, 0), w);
      u = normalize(u);      
      eye = eye + -g_cameraSpeed*u;
      g_camX = eye.x;
      g_camY = eye.y;
      g_camZ = eye.z;
      break;
    }
    /*
    case 'i':
      g_transX+=.01;
      printf("X: %f\n", g_transX);
      break;
    case 'o':
      g_transX-=.01;
      printf("X: %f\n", g_transX);
      break;
    case 'j':
      g_transY+=.01;
      printf("Y: %f\n", g_transY);
      break;
    case 'k':
      g_transY-=.01;
      printf("Y: %f\n", g_transY);
      break;
    case 'n':
      g_transZ+=.01;
      printf("Z: %f\n", g_transZ);
      break;
    case 'm':
      g_transZ-=.01;
      printf("Z: %f\n", g_transZ);
      break;
      */
    case 'p':
      g_showPath = !g_showPath;
      break;
    case '0':
      g_enableWind = !g_enableWind;
      break;
    case '1':
      g_wind.x += .5;
      break;
    case '2':
      g_wind.x -= .5;
      break;
    case '3':
      g_wind.z += .5;
      break;
    case '4':
      g_wind.z -= .5;
      break;
    case '[':
      g_brightness -= .01;
      break;
    case ']':
      g_brightness += .01;
      break;
      /*
    case 'r':
      g_resetTime = glutGet(GLUT_ELAPSED_TIME);
      animate(0);
      break;
      */
    case 'c':
      printf("Cam Data: <%f, %f, %f> looking %f, %f, %f at angle %f, %f\n",
              g_camX, g_camY, g_camZ, g_camLookX, g_camLookY, g_camLookZ, g_camTheta, g_camPhi); 
      break;
    case 'q': case 'Q' :
      exit( EXIT_SUCCESS );
      break;
  }
  glutPostRedisplay();
}

float p2wx(int in_x) {
  return ((2/(float)(g_width-1))*in_x)-1;
}

float p2wy(int in_y) {
  //flip glut y
  in_y = g_height - in_y;
  return ((2/(float)(g_height-1))*in_y)-1;
}

void mouse(int button, int state, int x, int y) { }

void mouseMove (int x, int y) {
   float currX = p2wx(x);
   float currY = p2wy(y);
   g_camTheta += .5 * currX;
   g_camPhi += .5 * currY;

   if (g_camPhi > 85*PI/180) { g_camPhi = 85*PI/180; }
   if (g_camPhi < -85*PI/180) {g_camPhi = -85*PI/180; }
   
   g_camLookX = cos(g_camPhi)*cos(g_camTheta);
   g_camLookY = sin(g_camPhi); 
   g_camLookZ = cos(g_camPhi)*cos(PI/2-g_camTheta);
}

int main( int argc, char *argv[]) {
 	glutInit( &argc, argv );
 	glutInitWindowPosition( 20, 20 );
 	glutInitWindowSize( 800, 600 );
 	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
 	glutCreateWindow("Cube and Transforms");

 	glutFullScreen();
 	glutSetCursor(GLUT_CURSOR_NONE); 
 	glutReshapeFunc( ReshapeGL );
 	glutDisplayFunc( Draw );
  glutKeyboardFunc( keyboard );
  glutMouseFunc( mouse );
  glutMotionFunc( mouseMove );
  glutPassiveMotionFunc( mouseMove );

  glutTimerFunc(100.0, animate, 0);

 	Initialize();
	

	//test the openGL version
	getGLversion();
	//install the shader
	if (!InstallShader(textFileRead((char *)"smooth_vert.glsl"), textFileRead((char *)"smooth_frag.glsl"))) {
		printf("Error installing shader!\n");
		return 0;
	}
		
	InitGeom();
  	glutMainLoop();
   	return 0;
}
