#version 150

struct Material {
  vec3 aColor;
  vec3 dColor;
  vec3 sColor;
  float alpha;
  float shine;
};

uniform Material uMat;

attribute vec3 aPosition;
attribute vec3 aNormal;

uniform vec3 uLightPos;
uniform vec3 uLightColor;
uniform vec3 uBrightness;

varying vec3 vNormal;
varying vec4 vPosition;
varying vec4 vWPosition;

void main() {
  vec4 Dl;
  vec4 Sl;
  vec4 normal = normalize(vec4(vNormal, 1.0));
  
  //Diffuse Lighting = Color_of_light * dot(vNormal, light_source) * color_of_object
  Dl = vec4(uLightPos, 1.0) - vPosition;
  Dl = normalize(Dl);
  Dl = vec4(uLightColor*(max(dot(normal, Dl), 0.0)*uMat.dColor), 0.0);

  //Specular Lighting = Color_of_light * dot(view_vector, reflected_vector)^shininess * Color_of_model_shine
  vec4 reflected = -1.0*vec4(uLightPos, 1.0) + ((2*max(dot(vec4(normalize(uLightPos), 1.0), normal), 0.0)) * normal);
  reflected = normalize(reflected);
  vec4 view = normalize(-1.0 * vPosition);
  Sl = vec4(uLightColor, 1.0) * pow(max(dot(view, reflected), 0.0), uMat.shine) * vec4(uMat.sColor, 0.0);
  //Sl = max(dot(view, reflected), 0.0) * vec4(1.0);
  
  gl_FragColor = Dl + Sl + vec4(uMat.aColor, 0.0) + vec4(uBrightness, uMat.alpha);
  //gl_FragColor = normalize(vec4(vNormal, 1));
    
}
